# pylint: disable=invalid-name,missing-docstring,protected-access
import pytest

import apypie
import requests
import json


def test_init(api):
    assert api
    assert api.apidoc


def test_init_bad_cachedir(tmpdir):
    bad_cachedir = tmpdir.join('bad')
    bad_cachedir.ensure(file=True)
    api = apypie.Api(uri='https://api.example.com', apidoc_cache_dir=bad_cachedir.strpath)
    with pytest.raises(OSError):
        api.apidoc


def test_init_bad_response(requests_mock, tmpdir):
    requests_mock.get('https://api.example.com/apidoc/v1.json', status_code=404)
    api = apypie.Api(uri='https://api.example.com', apidoc_cache_dir=tmpdir.strpath)
    with pytest.raises(apypie.exceptions.DocLoadingError) as excinfo:
        api.apidoc
    assert "Could not load data from https://api.example.com" in str(excinfo.value)


def test_init_with_lang(fixture_dir, requests_mock, tmpdir):
    with fixture_dir.join('dummy.json').open() as read_file:
        data = json.load(read_file)
    requests_mock.get('https://api.example.com/apidoc/v1.tlh.json', json=data)
    api = apypie.Api(uri='https://api.example.com', apidoc_cache_dir=tmpdir.strpath, language='tlh')
    assert api.apidoc


def test_init_with_missing_lang(fixture_dir, requests_mock, tmpdir):
    with fixture_dir.join('dummy.json').open() as read_file:
        data = json.load(read_file)
    requests_mock.get('https://api.example.com/apidoc/v1.tlh.json', status_code=404)
    requests_mock.get('https://api.example.com/apidoc/v1.json', json=data)
    api = apypie.Api(uri='https://api.example.com', apidoc_cache_dir=tmpdir.strpath, language='tlh')
    assert api.apidoc


def test_init_with_lang_family(fixture_dir, requests_mock, tmpdir):
    with fixture_dir.join('dummy.json').open() as read_file:
        data = json.load(read_file)
    requests_mock.get('https://api.example.com/apidoc/v1.tlh_EN.json', status_code=404)
    requests_mock.get('https://api.example.com/apidoc/v1.tlh.json', json=data)
    api = apypie.Api(uri='https://api.example.com', apidoc_cache_dir=tmpdir.strpath, language='tlh_EN')
    assert api.apidoc


def test_init_with_xdg_cachedir(fixture_dir, requests_mock, tmp_xdg_cache_home):
    with fixture_dir.join('dummy.json').open() as read_file:
        data = json.load(read_file)
    requests_mock.get('https://api.example.com/apidoc/v1.json', json=data)
    api = apypie.Api(uri='https://api.example.com')
    assert api.apidoc
    assert tmp_xdg_cache_home.join('apypie', 'https___api.example.com', 'v1', 'default.json').check(file=1)


def test_init_with_existing_cachedir(fixture_dir, requests_mock, tmp_xdg_cache_home):
    with fixture_dir.join('dummy.json').open() as read_file:
        data = json.load(read_file)
    requests_mock.get('https://api.example.com/apidoc/v1.json', json=data)

    json_path = tmp_xdg_cache_home.join('apypie', 'https___api.example.com', 'v1', 'deadc0ffee.json')
    json_path.ensure(file=True)
    fixture_dir.join('dummy.json').copy(json_path)

    api = apypie.Api(uri='https://api.example.com')
    api.apidoc

    assert json_path.check(file=1)
    assert tmp_xdg_cache_home.join('apypie', 'https___api.example.com', 'v1', 'default.json').check(exists=0)


def test_init_with_bad_cache(fixture_dir, requests_mock, tmp_xdg_cache_home):
    with fixture_dir.join('dummy.json').open() as read_file:
        data = json.load(read_file)
    requests_mock.get('https://api.example.com/apidoc/v1.json', json=data)

    json_path = tmp_xdg_cache_home.join('apypie', 'https___api.example.com', 'v1', 'deadc0ffee.json')
    json_path.ensure(file=True)
    json_path.write('BAD JSON')

    api = apypie.Api(uri='https://api.example.com')
    api.apidoc

    assert json_path.check(file=1)
    assert tmp_xdg_cache_home.join('apypie', 'https___api.example.com', 'v1', 'default.json').check(exists=0)


@pytest.mark.parametrize('username,password,expected', [
    (None, None, None),
    ('user', None, None),
    (None, 'pass', None),
    ('user', 'pass', ('user', 'pass')),
])
def test_init_auth(apidoc_cache_dir, username, password, expected):
    api = apypie.Api(uri='https://api.example.com', apidoc_cache_dir=apidoc_cache_dir.strpath,
                     username=username, password=password)

    assert api._session.auth == expected


@pytest.mark.parametrize('client_cert,client_key,expected', [
    (None, None, None),
    ('client.crt', None, None),
    (None, 'client.key', None),
    ('client.crt', 'client.key', ('client.crt', 'client.key')),
])
def test_init_cert(apidoc_cache_dir, client_cert, client_key, expected):
    api = apypie.Api(uri='https://api.example.com', apidoc_cache_dir=apidoc_cache_dir.strpath,
                     client_cert=client_cert, client_key=client_key)

    assert api._session.cert == expected


def test_init_language(apidoc_cache_dir):
    api = apypie.Api(uri='https://api.example.com', apidoc_cache_dir=apidoc_cache_dir.strpath,
                     language='tlh')

    assert api.language == 'tlh'
    assert api._session.headers['Accept-Language'] == 'tlh'


def test_resources(api):
    expected = sorted(['posts', 'users', 'comments'])
    assert expected == api.resources


def test_call_method(api, mocker):
    params = {'a': 1}
    headers = {'content-type': 'application/json'}
    mocker.patch('apypie.Api.http_call', autospec=True)
    api.call('users', 'index', params, headers)
    api.http_call.assert_called_once_with(api, 'get', '/users', params, headers, None, None)


def test_call_method_and_skip_validation(api, mocker):
    params = {'user': {'vip': True}}
    headers = {'content-type': 'application/json'}
    options = {'skip_validation': True}
    mocker.patch('apypie.Api.http_call', autospec=True)
    api.call('users', 'create', params, headers, options)
    api.http_call.assert_called_once_with(api, 'post', '/users', params, headers, None, None)


def test_call_method_and_fill_params(api, mocker):
    params = {'id': 1}
    headers = {'content-type': 'application/json'}
    mocker.patch('apypie.Api.http_call', autospec=True)
    api.call('users', 'show', params, headers)
    api.http_call.assert_called_once_with(api, 'get', '/users/1', {}, headers, None, None)


def test_http_call_get(api, requests_mock):
    requests_mock.get('https://api.example.com/', text='{}')
    api.http_call('get', '/')


def test_http_call_get_querystring(api, requests_mock):
    requests_mock.get('https://api.example.com/?flag=true', text='{}')
    api.http_call('get', '/', params={'flag': True})


def test_http_call_get_headers(api, requests_mock):
    headers = {'X-Apypie-Test': 'Awesome'}
    expected_headers = {
        'Accept': 'application/json;version=1',
        'Accept-Encoding': 'gzip, deflate',
        'User-Agent': 'apypie (https://github.com/Apipie/apypie)',
    }
    expected_headers.update(headers)
    requests_mock.get('https://api.example.com/', request_headers=expected_headers, text='{}')
    api.http_call('get', '/', headers=headers)


def test_http_call_get_headers_auth(apidoc_cache_dir, requests_mock):
    headers = {'X-Apypie-Test': 'Awesome'}
    expected_headers = {
        'Accept': 'application/json;version=1',
        'Authorization': 'Basic dXNlcjpwYXNz',
    }
    expected_headers.update(headers)
    requests_mock.get('https://api.example.com/', request_headers=expected_headers, text='{}')
    api = apypie.Api(uri='https://api.example.com', apidoc_cache_dir=apidoc_cache_dir.strpath,
                     username='user', password='pass')
    api.http_call('get', '/', headers=headers)


def test_http_call_get_headers_auth_lang(apidoc_cache_dir, requests_mock):
    headers = {'X-Apypie-Test': 'Awesome'}
    expected_headers = {
        'Accept': 'application/json;version=1',
        'Accept-Language': 'tlh',
        'Authorization': 'Basic dXNlcjpwYXNz',
    }
    expected_headers.update(headers)
    requests_mock.get('https://api.example.com/', request_headers=expected_headers, text='{}')
    api = apypie.Api(uri='https://api.example.com', apidoc_cache_dir=apidoc_cache_dir.strpath,
                     username='user', password='pass', language='tlh')
    api.http_call('get', '/', headers=headers)


def test_http_call_get_headers_lang(apidoc_cache_dir, requests_mock):
    headers = {'X-Apypie-Test': 'Awesome'}
    expected_headers = {
        'Accept': 'application/json;version=1',
        'Accept-Language': 'tlh',
    }
    expected_headers.update(headers)
    requests_mock.get('https://api.example.com/', request_headers=expected_headers, text='{}')
    api = apypie.Api(uri='https://api.example.com', apidoc_cache_dir=apidoc_cache_dir.strpath,
                     language='tlh')
    api.http_call('get', '/', headers=headers)


def test_http_call_get_headers_cache(fixture_dir, requests_mock, tmp_xdg_cache_home):
    response_headers = {'Apipie-Checksum': 'c0ffeec0ffee'}
    with fixture_dir.join('dummy.json').open() as read_file:
        data = json.load(read_file)
    requests_mock.get('https://api.example.com/apidoc/v1.json', headers=response_headers, json=data)

    api = apypie.Api(uri='https://api.example.com')
    api.apidoc

    print(api.apidoc_cache_name)

    assert tmp_xdg_cache_home.join('apypie', 'https___api.example.com', 'v1', 'default.json').check(exists=0)
    assert tmp_xdg_cache_home.join('apypie', 'https___api.example.com', 'v1', 'c0ffeec0ffee.json').check(file=1)


def test_http_call_get_with_params(api, requests_mock):
    requests_mock.get('https://api.example.com/?test=all+the+things', text='{}')
    api.http_call('get', '/', {'test': 'all the things'})


def test_http_call_post(api, requests_mock):
    expected_headers = {
        'Accept': 'application/json;version=1',
        'Content-Type': 'application/json',
    }
    requests_mock.post('https://api.example.com/', text='{}', request_headers=expected_headers)
    api.http_call('post', '/')


def test_http_call_post_with_params(api, requests_mock):
    expected_headers = {
        'Accept': 'application/json;version=1',
        'Content-Type': 'application/json',
    }
    params = {'test': 'all the things'}
    requests_mock.post('https://api.example.com/', text='{}', request_headers=expected_headers)
    api.http_call('post', '/', params)


def test_http_call_put(api, requests_mock):
    expected_headers = {
        'Accept': 'application/json;version=1',
        'Content-Type': 'application/json',
    }
    requests_mock.put('https://api.example.com/', text='{}', request_headers=expected_headers)
    api.http_call('put', '/')


def test_http_call_put_with_params(api, requests_mock):
    expected_headers = {
        'Accept': 'application/json;version=1',
        'Content-Type': 'application/json',
    }
    params = {'test': 'all the things'}
    requests_mock.put('https://api.example.com/', text='{}', request_headers=expected_headers)
    api.http_call('put', '/', params)


def test_http_call_delete(api, requests_mock):
    requests_mock.delete('https://api.example.com/', text='{}')
    api.http_call('delete', '/')


def test_http_call_delete_with_params(api, requests_mock):
    expected_headers = {
        'Accept': 'application/json;version=1',
        'Content-Type': 'application/json',
    }
    requests_mock.delete('https://api.example.com/', text='{}', request_headers=expected_headers)
    api.http_call('delete', '/', {'test': 'all the things'})


def test_http_call_with_no_content_answer(api, requests_mock):
    requests_mock.delete('https://api.example.com/', status_code=204)
    api.http_call('delete', '/', {})


def test_clean_cache(fixture_dir, requests_mock, tmp_xdg_cache_home):
    with fixture_dir.join('dummy.json').open() as read_file:
        data = json.load(read_file)
    requests_mock.get('https://api.example.com/apidoc/v1.json', json=data)
    api = apypie.Api(uri='https://api.example.com')
    assert api.apidoc
    assert tmp_xdg_cache_home.join('apypie', 'https___api.example.com', 'v1', 'default.json').check(file=1)
    api.clean_cache()
    assert api._apidoc is None
    assert tmp_xdg_cache_home.join('apypie', 'https___api.example.com', 'v1', 'default.json').check(exists=0)


def test_validate_cache(fixture_dir, requests_mock, tmp_xdg_cache_home):
    with fixture_dir.join('dummy.json').open() as read_file:
        data = json.load(read_file)
    requests_mock.get('https://api.example.com/apidoc/v1.json', json=data)
    api = apypie.Api(uri='https://api.example.com')
    assert api.apidoc
    assert tmp_xdg_cache_home.join('apypie', 'https___api.example.com', 'v1', 'default.json').check(file=1)
    api.validate_cache('testcache')
    assert tmp_xdg_cache_home.join('apypie', 'https___api.example.com', 'v1', 'default.json').check(exists=0)
    assert api.apidoc
    assert tmp_xdg_cache_home.join('apypie', 'https___api.example.com', 'v1', 'testcache.json').check(file=1)


def test_validate_cache_path_traversal(fixture_dir, requests_mock, tmp_xdg_cache_home):
    with fixture_dir.join('dummy.json').open() as read_file:
        data = json.load(read_file)
    requests_mock.get('https://api.example.com/apidoc/v1.json', json=data)
    api = apypie.Api(uri='https://api.example.com')
    assert api.apidoc
    assert tmp_xdg_cache_home.join('apypie', 'https___api.example.com', 'v1', 'default.json').check(file=1)
    api.validate_cache('../help/testcache')
    assert tmp_xdg_cache_home.join('apypie', 'https___api.example.com', 'v1', 'default.json').check(exists=0)
    assert api.apidoc
    assert tmp_xdg_cache_home.join('apypie', 'https___api.example.com', 'v1', 'testcache.json').check(file=1)


def test_custom_session(fixture_dir, requests_mock, tmpdir):
    headers = {'X-Apypie-Test': 'Custom'}

    my_session = requests.Session()
    my_session.headers = headers

    my_api = apypie.Api(uri='https://api.example.com', session=my_session)

    requests_mock.get('https://api.example.com/', request_headers=headers, text='{}')
    my_api.http_call('get', '/')



from __future__ import print_function

import pytest
import apypie


def test_docloadingerror(capsys):
    test_message = 'this is DocLoadingError'
    test_error = apypie.exceptions.DocLoadingError(test_message)
    with pytest.raises(apypie.exceptions.DocLoadingError) as e:
        raise test_error
    print(e.value)
    captured = capsys.readouterr()
    assert captured.out == '{}\n'.format(test_message)



# pylint: disable=invalid-name,missing-docstring,protected-access
import io
import json

import pytest
import requests
import requests.exceptions

from apypie.foreman import ForemanApi, ForemanApiException, _recursive_dict_keys


def test_recursive_dict_keys():
    a_dict = {'level1': 'has value', 'level2': {'real_level2': 'more value', 'level3': {'real_level3': 'nope'}}}
    expected_keys = set(['level1', 'level2', 'level3', 'real_level2', 'real_level3'])
    assert _recursive_dict_keys(a_dict) == expected_keys


@pytest.fixture
def foremanapi(fixture_dir, requests_mock, tmpdir):
    with fixture_dir.join('foreman.json').open() as read_file:
        data = json.load(read_file)
    requests_mock.get('https://api.example.com/apidoc/v2.json', json=data)
    return ForemanApi(uri='https://api.example.com', apidoc_cache_dir=tmpdir.strpath)


@pytest.fixture
def lunaapi(fixture_dir, requests_mock, tmpdir):
    with fixture_dir.join('luna.json').open() as read_file:
        data = json.load(read_file)
    requests_mock.get('https://api.example.com/apidoc/v2.json', json=data)
    return ForemanApi(uri='https://api.example.com', apidoc_cache_dir=tmpdir.strpath)


def test_init(foremanapi):
    assert foremanapi
    assert foremanapi.apidoc


def test_resources(foremanapi):
    assert 'domains' in foremanapi.resources


def test_resource_action(foremanapi, requests_mock):
    requests_mock.get('https://api.example.com/api/organizations/1', json={'id': 1})
    org = foremanapi.resource_action('organizations', 'show', {'id': 1})
    assert org


def test_resource_action_unknown_resource(foremanapi):
    with pytest.raises(ForemanApiException) as excinfo:
        foremanapi.resource_action('bubblegums', 'show', {'id': 1})
    assert "The server doesn't know about bubblegums, is the right plugin installed?" in str(excinfo.value)


def test_resource_action_http_error(foremanapi, mocker):
    raw_bytes = io.BytesIO(b'This is bad')
    response = requests.Response()
    response.raw = raw_bytes
    exception = requests.exceptions.HTTPError(response=response)
    mocker.patch('apypie.foreman.ForemanApi._resource_call', autospec=True, side_effect=exception)
    with pytest.raises(ForemanApiException) as excinfo:
        foremanapi.resource_action('organizations', 'show', {'id': 1})
    assert "Error while performing show on organizations:  - This is bad" in str(excinfo.value)


def test_resource_action_http_error_json(foremanapi, mocker):
    raw_bytes = io.BytesIO(b'{"error":"superbad"}')
    response = requests.Response()
    response.raw = raw_bytes
    exception = requests.exceptions.HTTPError(response=response)
    mocker.patch('apypie.foreman.ForemanApi._resource_call', autospec=True, side_effect=exception)
    with pytest.raises(ForemanApiException) as excinfo:
        foremanapi.resource_action('organizations', 'show', {'id': 1})
    assert "Error while performing show on organizations:  - superbad" in str(excinfo.value)


def test_resource_action_http_error_other_json(foremanapi, mocker):
    raw_bytes = io.BytesIO(b'{"message":"E_TOO_BAD"}')
    response = requests.Response()
    response.raw = raw_bytes
    exception = requests.exceptions.HTTPError(response=response)
    mocker.patch('apypie.foreman.ForemanApi._resource_call', autospec=True, side_effect=exception)
    with pytest.raises(ForemanApiException) as excinfo:
        foremanapi.resource_action('organizations', 'show', {'id': 1})
    assert "Error while performing show on organizations:  - {'message': 'E_TOO_BAD'}" in str(excinfo.value)


def test_resource_action_wait_for_task(lunaapi, requests_mock):
    running_task = {'id': 1, 'state': 'running', 'action': 'test', 'started_at': 'now'}
    stopped_task = {'id': 1, 'state': 'stopped', 'result': 'success'}
    requests_mock.post('https://api.example.com/api/domains', json=running_task)
    requests_mock.get('https://api.example.com/foreman_tasks/api/tasks/1', json=stopped_task)
    done = lunaapi.resource_action('domains', 'create', {'name': 'test'})
    assert done['result'] == 'success'


def test_wait_for_task(lunaapi, requests_mock):
    running_task = {'id': 1, 'state': 'running'}
    stopped_task = {'id': 1, 'state': 'stopped', 'result': 'success'}
    requests_mock.get('https://api.example.com/foreman_tasks/api/tasks/1', json=stopped_task)
    lunaapi.wait_for_task(running_task)


def test_wait_for_task_failed_task(lunaapi, requests_mock):
    running_task = {'id': 1, 'state': 'running'}
    stopped_task = {'id': 1, 'state': 'stopped', 'result': 'error', 'action': 'test', 'humanized': {'errors': 'you lost the game'}}
    requests_mock.get('https://api.example.com/foreman_tasks/api/tasks/1', json=stopped_task)
    with pytest.raises(ForemanApiException) as excinfo:
        lunaapi.wait_for_task(running_task)
    assert "Task test(1) did not succeed. Task information: you lost the game" in str(excinfo.value)


def test_wait_for_task_timeout(lunaapi, requests_mock):
    running_task = {'id': 1, 'state': 'running'}
    requests_mock.get('https://api.example.com/foreman_tasks/api/tasks/1', json=running_task)
    lunaapi.task_timeout = 1
    with pytest.raises(ForemanApiException) as excinfo:
        lunaapi.wait_for_task(running_task)
    assert "Timeout waiting for Task 1" in str(excinfo.value)


def test_show(foremanapi, requests_mock):
    requests_mock.get('https://api.example.com/api/organizations/1', json={'id': 1})
    org = foremanapi.show('organizations', 1)
    assert org


def test_list(foremanapi, requests_mock):
    requests_mock.get('https://api.example.com/api/organizations?per_page=4294967296', complete_qs=True, json={'results': [{'id': 1}]})
    orgs = foremanapi.list('organizations')
    assert orgs


def test_list_with_search(foremanapi, requests_mock):
    requests_mock.get('https://api.example.com/api/organizations?search=name%3DTEST&per_page=4294967296', complete_qs=True, json={'results': [{'id': 1}]})
    orgs = foremanapi.list('organizations', search='name=TEST')
    assert orgs


def test_create(foremanapi, requests_mock):
    def match_json_body(request):
        return {'organization': {'name': 'Test'}} == request.json()

    requests_mock.post('https://api.example.com/api/organizations', additional_matcher=match_json_body, json={'id': 1})
    org = foremanapi.create('organizations', {'name': 'Test'})
    assert org


def test_update(foremanapi, requests_mock):
    def match_json_body(request):
        return {'organization': {'name': 'Test'}} == request.json()

    requests_mock.put('https://api.example.com/api/organizations/1', additional_matcher=match_json_body, json={'id': 1})
    org = foremanapi.update('organizations', {'id': 1, 'name': 'Test'})
    assert org


def test_delete(foremanapi, requests_mock):
    requests_mock.delete('https://api.example.com/api/organizations/1', status_code=204)
    foremanapi.delete('organizations', {'id': 1})


@pytest.mark.parametrize("params,expected", [
    ({'name': 'test'}, ({'organization': {'name': 'test'}}, set())),
    ({'name': 'test', 'nope': 'nope'}, ({'organization': {'name': 'test'}}, {'nope'})),
])
def test_validate_payload(foremanapi, params, expected):
    result = foremanapi.validate_payload('organizations', 'create', params)
    assert result == expected



import apypie
import pytest


def test_resource_action(action):
    assert 'show' == action.name


def test_action_apidoc(action):
    assert action.apidoc


def test_action_routes(action):
    assert action.routes
    assert isinstance(action.routes[0], apypie.Route)


def test_action_route(action):
    route = action.routes[0]
    assert '/users/:id' == route.path
    assert 'get' == route.method


def test_action_params(resource):
    action = resource.action('create')
    assert action.params
    assert ['user'] == [param.name for param in action.params]


def test_action_examples(resource):
    action = resource.action('index')
    assert action.examples


def test_action_example(resource):
    action = resource.action('index')
    example = action.examples[0]
    assert 'GET' == example.http_method
    assert '/users' == example.path
    assert '' == example.args
    assert 200 == example.status
    assert '[ {"user":{"name":"John Doe" }} ]' in example.response


def test_action_call(resource, mocker):
    params = {}
    headers = {'content-type': 'application/json'}
    mocker.patch('apypie.Api.call', autospec=True)
    assert resource.action('index').call(params, headers)
    resource.api.call.assert_called_once_with(resource.api, resource.name, 'index', params, headers, None, None, None)


def test_action_find_route(resource):
    action = resource.action('index')
    assert '/users' == action.find_route().path


def test_action_find_route_ignoring_unknown_keys(resource):
    action = resource.action('index')
    params = {'per_page': 5}
    assert '/users' == action.find_route(params).path


def test_action_find_route_longest(api):
    action = api.resource('comments').action('archive')
    params = {'id': 1, 'user_id': 1}
    assert '/archive/users/:user_id/comments/:id' == action.find_route(params).path


def test_action_find_route_longest_ignoring_unknown_keys(api):
    action = api.resource('comments').action('archive')
    params = {'id': 1, 'user_id': 1, 'per_page': 5}
    assert '/archive/users/:user_id/comments/:id' == action.find_route(params).path


def test_action_find_route_longest_ignoring_none(api):
    action = api.resource('comments').action('archive')
    params = {'id': 1, 'user_id': None}
    assert '/archive/comments/:id' == action.find_route(params).path


def test_action_validate_missing_required_params(resource):
    action = resource.action('create')
    with pytest.raises(apypie.exceptions.MissingArgumentsError) as excinfo:
        action.validate({'user': {'vip': True}})
    assert "The following required parameters are missing" in str(excinfo.value)
    assert "name" in str(excinfo.value)


def test_action_validate_missing_nested_required_params(resource):
    action = resource.action('create')
    with pytest.raises(apypie.exceptions.MissingArgumentsError) as excinfo:
        action.validate({'user': {'name': 'John Doe', 'address': {'street': 'K JZD'}}})
    assert "The following required parameters are missing" in str(excinfo.value)
    assert "user[address][city]" in str(excinfo.value)


def test_action_validate_missing_nested_required_params_array(resource):
    action = resource.action('create')
    with pytest.raises(apypie.exceptions.MissingArgumentsError) as excinfo:
        action.validate({'user': {'name': 'John Doe', 'contacts': [{'kind': 'email'}]}})
    assert "The following required parameters are missing" in str(excinfo.value)
    assert "user[contacts][0][contact]" in str(excinfo.value)


def test_action_validate_missing_nested_invalid_params(resource):
    action = resource.action('create')
    with pytest.raises(apypie.exceptions.InvalidArgumentTypesError):
        action.validate({'user': {'name': 'John Doe', 'contacts': [1, 2]}})


def test_action_validate_invalid_boolean(resource):
    action = resource.action('create')
    with pytest.raises(ValueError) as excinfo:
        action.validate({'user': {'name': 'John Doe', 'vip': 'maybe'}})
    assert "vip (maybe): Must be one of: <code>true</code>, <code>false</code>, <code>1</code>, <code>0</code>" in str(excinfo.value)


def test_action_validate_valid_boolean(resource):
    action = resource.action('create')
    action.validate({'user': {'name': 'John Doe', 'vip': False}})


def test_action_validate_valid_boolean_as_int(resource):
    action = resource.action('create')
    action.validate({'user': {'name': 'John Doe', 'vip': 0}})


def test_action_validate_invalid_numeric(api):
    action = api.resource('comments').action('archive')
    with pytest.raises(ValueError) as excinfo:
        action.validate({'id': 'MYNUMBER'})
    assert "id (MYNUMBER): Must be a Integer" in str(excinfo.value)


def test_action_validate_valid_numeric(api):
    action = api.resource('comments').action('archive')
    action.validate({'id': 1})


def test_action_validate_valid_numeric_as_string(api):
    action = api.resource('comments').action('archive')
    action.validate({'id': '1'})


def test_action_validate_invalid_string(resource):
    action = resource.action('create')
    with pytest.raises(ValueError) as excinfo:
        action.validate({'user': {'name': []}})
    assert "name ([]): Must be a String" in str(excinfo.value)


def test_action_validate_valid_string(resource):
    action = resource.action('create')
    action.validate({'user': {'name': 'John Doe'}})


def test_action_validate_valid_string_as_int(resource):
    # workaround for https://projects.theforeman.org/issues/27107
    action = resource.action('create')
    action.validate({'user': {'name': 1}})


def test_action_validate_valid_none(foreman_api):
    action = foreman_api.resource('domains').action('update')
    action.validate({'id': 1, 'domain': {'dns_id': None}})


def test_action_validate_invalid_none(resource):
    action = resource.action('create')
    with pytest.raises(ValueError) as excinfo:
        action.validate({'user': {'name': None}})
    assert "name can't be None" in str(excinfo.value)


def test_action_validate_files(resource):
    # this pretends to upload the user as a file
    resource.action('create_unnested').validate({}, files={'name': 'John Doe'})


def test_action_validate_data(resource):
    # this pretends to upload the user as data
    resource.action('create_unnested').validate({}, data={'name': 'John Doe'})


def test_action_validate_minimal_correct_params(resource):
    resource.action('create').validate({'user': {'name': 'John Doe'}})
    resource.action('create_unnested').validate({'name': 'John Doe'})


def test_action_validate_full_correct_params(resource):
    action = resource.action('create')
    params = {'user': {
        'name': 'John Doe',
        'vip': True,
        'address': {
            'city': 'Ankh-Morpork',
            'street': 'Audit Alley'
        },
        'contacts': [
            {'contact': 'john@doe.org', 'kind': 'email'},
            {'contact': '123456', 'kind': 'pobox'}
        ]
    }}
    action.validate(params)


@pytest.mark.parametrize('resource,action,input_dict,expected_params', [
    ('comments', 'archive', {'id': 1}, {'id': 1}),
    ('comments', 'archive', {'id': 1, 'should_be_ignored': True}, {'id': 1}),
    ('comments', 'show', {'id': 1, 'post_id': 13, 'user_id': 666}, {'id': 1, 'post_id': 13, 'user_id': 666}),
    ('users', 'create', {'name': 'testuser'}, {'user': {'name': 'testuser'}}),
])
def test_action_prepare_params(api, resource, action, input_dict, expected_params):
    generated_params = api.resource(resource).action(action).prepare_params(input_dict)
    assert expected_params == generated_params


def test_action_prepare_params_hash_no_nested_params(foreman_api):
    # some actions have an expected_type of hash, but don't actually define how the hash should look like
    action = foreman_api.resource('compute_attributes').action('create')
    input_dict = {'compute_profile_id': 1, 'compute_resource_id': 1, 'vm_attrs': {'some_attr': 'value'}}
    expected_params = {'compute_profile_id': 1, 'compute_resource_id': 1, 'compute_attribute': {'vm_attrs': {'some_attr': 'value'}}}
    generated_params = action.prepare_params(input_dict)
    assert expected_params == generated_params


def test_action_prepare_params_hash_of_hashes(luna_api):
    action = luna_api.resource('job_invocations').action('create')
    input_dict = {'bookmark_id': 10, 'job_template_id': 177, 'recurrence': {'cron_line': '30 2 * * *'}, 'concurrency_control': {'concurrency_level': 2}, 'targeting_type': 'static_query', 'inputs': {'command': 'pwd'}}
    expected_params = {'job_invocation': {'bookmark_id': 10, 'job_template_id': 177, 'recurrence': {'cron_line': '30 2 * * *'}, 'concurrency_control': {'concurrency_level': 2}, 'targeting_type': 'static_query', 'inputs': {'command': 'pwd'}}}
    generated_params = action.prepare_params(input_dict)
    assert expected_params == generated_params



import pytest
import apypie


@pytest.fixture
def param():
    param = apypie.Param(
        allow_nil=False,
        description="<p>Architecture</p>",
        expected_type="hash",
        full_name="architecture",
        name="architecture",
        params=[
            {
                'allow_nil': False,
                'description': "",
                'expected_type': "string",
                'full_name': "architecture[name]",
                'name': "name",
                'required': False,
                'validator': "Must be String"
            }
        ],
        required=True,
        validator="Must be a Hash"
    )
    return param


def test_param(param):
    assert param


def test_param_nested_params(param):
    assert 'name' == param.params[0].name


def test_param_expected_type(param):
    assert "hash" == param.expected_type


def test_param_description_strip_html(param):
    assert 'Architecture' == param.description


def test_param_required(param):
    assert param.required
    assert not param.params[0].required


def test_param_validator(param):
    assert "Must be a Hash" == param.validator



import apypie


def test_example_apidoc_post():
    example_string = "POST /api/architectures\n{\n  \"architecture\": {\n    \"name\": \"i386\"\n  }\n}\n200\n{\n  \"architecture\": {\n    \"name\": \"i386\",\n    \"id\": 501905020,\n    \"updated_at\": \"2012-12-18T15:24:43Z\",\n    \"operatingsystem_ids\": [],\n    \"created_at\": \"2012-12-18T15:24:43Z\"\n  }\n}"
    example = apypie.Example.parse(example_string)
    assert 'POST' == example.http_method
    assert '/api/architectures' == example.path
    assert '{\n  "architecture": {\n    "name": "i386"\n  }\n}' == example.args
    assert 200 == example.status
    assert '{\n  "architecture": ' in example.response


def test_example_apidoc_get():
    example_string = "GET /api/architectures/x86_64\n200\n{\n  \"architecture\": {\n    \"name\": \"x86_64\",\n    \"id\": 501905019,\n    \"updated_at\": \"2012-12-18T15:24:42Z\",\n    \"operatingsystem_ids\": [\n      309172073,\n      1073012828,\n      331303656\n    ],\n    \"created_at\": \"2012-12-18T15:24:42Z\"\n  }\n}"
    example = apypie.Example.parse(example_string)
    assert 'GET' == example.http_method
    assert '/api/architectures/x86_64' == example.path
    assert '' == example.args
    assert 200 == example.status
    assert '{\n  "architecture": ' in example.response



"""
Apypie Foreman module

opinionated helpers to use Apypie with Foreman
"""
import time

from typing import cast, Optional, Set, Tuple

from apypie.api import Api

from apypie.resource import Resource  # pylint: disable=unused-import  # noqa: F401

# Foreman supports "per_page=all" since 2.2 (https://projects.theforeman.org/issues/29909)
# But plugins, especially Katello, do not: https://github.com/Katello/katello/pull/11126
# To still be able to fetch all results without pagination, we have this constant for now
PER_PAGE = 2 << 31


class ForemanApiException(Exception):
    """
    General Exception, raised by any issue in ForemanApi
    """

    def __init__(self, msg, error=None):
        if error:
            super().__init__(f'{msg} - {error}')
        else:
            super().__init__(msg)

    @classmethod
    def from_exception(cls, exc, msg):
        """
        Create a ForemanException from any other Exception

        Especially useful to gather the error message from HTTP responses.
        """
        error = None
        if hasattr(exc, 'response') and exc.response is not None:
            try:
                response = exc.response.json()
                if 'error' in response:
                    error = response['error']
                else:
                    error = response
            except Exception:  # pylint: disable=broad-except
                error = exc.response.text
        return cls(msg=msg, error=error)


class ForemanApi(Api):
    """
    `apypie.Api` with default settings and helper functions for Foreman

    Usage::

      >>> import apypie
      >>> api = apypie.ForemanApi(uri='https://foreman.example.com', username='admin', password='changeme')
    """

    def __init__(self, **kwargs):
        self.task_timeout = kwargs.pop('task_timeout', 60)
        self.task_poll = 4
        kwargs['api_version'] = 2
        super().__init__(**kwargs)

    def _resource(self, resource: str) -> 'Resource':
        if resource not in self.resources:
            raise ForemanApiException(msg=f"The server doesn't know about {resource}, is the right plugin installed?")
        return self.resource(resource)

    def _resource_call(self, resource: str, *args, **kwargs) -> Optional[dict]:
        return self._resource(resource).call(*args, **kwargs)

    def _resource_prepare_params(self, resource: str, action: str, params: dict) -> dict:
        api_action = self._resource(resource).action(action)
        return api_action.prepare_params(params)

    def resource_action(self, resource: str, action: str, params: dict, options=None, data=None, files=None,  # pylint: disable=too-many-arguments
                        ignore_task_errors: bool = False) -> Optional[dict]:
        """
        Perform a generic action on a resource

        Will wait for tasks if the action returns one
        """
        resource_payload = self._resource_prepare_params(resource, action, params)
        if options is None:
            options = {}
        try:
            result = self._resource_call(resource, action, resource_payload, options=options, data=data, files=files)
            is_foreman_task = isinstance(result, dict) and 'action' in result and 'state' in result and 'started_at' in result
            if result and is_foreman_task:
                result = self.wait_for_task(result, ignore_errors=ignore_task_errors)
        except Exception as exc:
            msg = f'Error while performing {action} on {resource}: {exc}'
            raise ForemanApiException.from_exception(exc, msg) from exc
        return result

    def wait_for_task(self, task: dict, ignore_errors: bool = False) -> dict:
        """
        Wait for a foreman-tasks task, polling it every ``self.task_poll`` seconds.

        Will raise a ForemanApiException when task has not finished in ``self.task_timeout`` seconds.
        """
        duration = self.task_timeout
        while task['state'] not in ['paused', 'stopped']:
            duration -= self.task_poll
            if duration <= 0:
                raise ForemanApiException(msg=f"Timeout waiting for Task {task['id']}")
            time.sleep(self.task_poll)

            resource_payload = self._resource_prepare_params('foreman_tasks', 'show', {'id': task['id']})
            task = cast(dict, self._resource_call('foreman_tasks', 'show', resource_payload))
        if not ignore_errors and task['result'] != 'success':
            msg = f"Task {task['action']}({task['id']}) did not succeed. Task information: {task['humanized']['errors']}"
            raise ForemanApiException(msg=msg)
        return task

    def show(self, resource: str, resource_id: int, params: Optional[dict] = None) -> Optional[dict]:
        """
        Execute the ``show`` action on an entity.

        :param resource: Plural name of the api resource to show
        :param resource_id: The ID of the entity to show
        :param params: Lookup parameters (i.e. parent_id for nested entities)

        :return: The entity
        """
        payload = {'id': resource_id}
        if params:
            payload.update(params)
        return self.resource_action(resource, 'show', payload)

    def list(self, resource: str, search: Optional[str] = None, params: Optional[dict] = None) -> list:
        """
        Execute the ``index`` action on an resource.

        :param resource: Plural name of the api resource to show
        :param search: Search string as accepted by the API to limit the results
        :param params: Lookup parameters (i.e. parent_id for nested entities)

        :return: List of results
        """
        payload: dict = {'per_page': PER_PAGE}
        if search is not None:
            payload['search'] = search
        if params:
            payload.update(params)

        result = self.resource_action(resource, 'index', payload)
        if result:
            return result['results']
        return []

    def create(self, resource: str, desired_entity: dict, params: Optional[dict] = None) -> Optional[dict]:
        """
        Create entity with given properties

        :param resource: Plural name of the api resource to manipulate
        :param desired_entity: Desired properties of the entity
        :param params: Lookup parameters (i.e. parent_id for nested entities)

        :return: The new current state of the entity
        """
        payload = desired_entity.copy()
        if params:
            payload.update(params)
        return self.resource_action(resource, 'create', payload)

    def update(self, resource: str, desired_entity: dict, params: Optional[dict] = None) -> Optional[dict]:
        """
        Update entity with given properties

        :param resource: Plural name of the api resource to manipulate
        :param desired_entity: Desired properties of the entity
        :param params: Lookup parameters (i.e. parent_id for nested entities)

        :return: The new current state of the entity
        """
        payload = desired_entity.copy()
        if params:
            payload.update(params)
        return self.resource_action(resource, 'update', payload)

    def delete(self, resource: str, current_entity: dict, params: Optional[dict] = None) -> None:
        """
        Delete a given entity

        :param resource: Plural name of the api resource to manipulate
        :param current_entity: Current properties of the entity
        :param params: Lookup parameters (i.e. parent_id for nested entities)

        :return: The new current state of the entity
        """
        payload = {'id': current_entity['id']}
        if params:
            payload.update(params)
        entity = self.resource_action(resource, 'destroy', payload)

        # this is a workaround for https://projects.theforeman.org/issues/26937
        if entity and isinstance(entity, dict) and 'error' in entity and 'message' in entity['error']:
            raise ForemanApiException(msg=entity['error']['message'])

    def validate_payload(self, resource: str, action: str, payload: dict) -> Tuple[dict, Set[str]]:
        """
        Check whether the payload only contains supported keys.

        :param resource: Plural name of the api resource to check
        :param action: Name of the action to check payload against
        :param payload: API paylod to be checked

        :return: The payload as it can be submitted to the API and set of unssuported parameters
        """
        filtered_payload = self._resource_prepare_params(resource, action, payload)
        unsupported_parameters = set(payload.keys()) - _recursive_dict_keys(filtered_payload)
        return (filtered_payload, unsupported_parameters)


def _recursive_dict_keys(a_dict: dict) -> set:
    """Find all keys of a nested dictionary"""
    keys = set(a_dict.keys())
    for value in a_dict.values():
        if isinstance(value, dict):
            keys.update(_recursive_dict_keys(value))
    return keys



import pytest
import apypie


def test_resource_actions(resource):
    expected = sorted(['index', 'show', 'create', 'update', 'destroy', 'create_unnested'])
    assert expected == resource.actions


def test_resource_has_action(resource):
    assert resource.has_action('index')


def test_resource_has_action_false(resource):
    assert not resource.has_action('missing')


def test_resource_action_missing(resource):
    with pytest.raises(KeyError) as excinfo:
        resource.action('missing')
    assert "Unknown action 'missing'" in str(excinfo.value)
    assert "Supported actions: create, create_unnested, destroy, index, show, update" in str(excinfo.value)


def test_resource_action_type(resource):
    assert isinstance(resource.action('index'), apypie.Action)


def test_resource_call_action(resource, mocker):
    params = {}
    headers = {'content-type': 'application/json'}
    mocker.patch('apypie.Api.call', autospec=True)
    assert resource.call('index', params, headers)
    resource.api.call.assert_called_once_with(resource.api, resource.name, 'index', params, headers, None, None, None)


def test_resource_call_action_minimal(resource, mocker):
    mocker.patch('apypie.Api.call', autospec=True)
    assert resource.call('index')
    resource.api.call.assert_called_once_with(resource.api, resource.name, 'index', None, None, None, None, None)


def test_resource_existing(resource):
    assert 'users' == resource.name


def test_resource_missing(api):
    with pytest.raises(KeyError) as excinfo:
        api.resource('missing')
    assert "Resource 'missing' does not exist in the API." in str(excinfo.value)
    assert "Existing resources: comments, posts, users" in str(excinfo.value)



"""
Apypie - Apipie bindings for Python
"""

from __future__ import print_function, absolute_import

from apypie.resource import Resource
from apypie.action import Action
from apypie.route import Route
from apypie.api import Api
from apypie.example import Example
from apypie.param import Param
from apypie.inflector import Inflector
from apypie.foreman import ForemanApi, ForemanApiException

__all__ = ['Api', 'Resource', 'Route', 'Action', 'Example', 'Param', 'Inflector', 'ForemanApi', 'ForemanApiException']



import pytest
import apypie


@pytest.fixture
def route():
    return apypie.Route("/api/architectures/:id", "GET")


def test_route(route):
    assert route


def test_route_params_in_path(route):
    assert ['id'] == route.params_in_path


def test_route_method_lower(route):
    assert 'get' == route.method


def test_route_path_with_params_fill(route):
    assert '/api/architectures/1' == route.path_with_params({'id': 1})


def test_route_path_with_escaped_params(route):
    assert r'/api/architectures/nested%2Fid' == route.path_with_params({'id': 'nested/id'})


def test_route_path_with_params_fill_wrong(route):
    with pytest.raises(KeyError) as excinfo:
        route.path_with_params({'wrong': 1})
    assert "missing param 'id' in parameters" in str(excinfo.value)


def test_route_path_with_params(route):
    assert '/api/architectures/:id' == route.path_with_params()



from __future__ import print_function

import apypie


def test_inflector():
    inflector = apypie.Inflector()
    assert inflector.pluralize('word') == 'words'
    assert inflector.singularize('words') == 'word'
    assert inflector.pluralize('gpg_key') == 'gpg_keys'
    assert inflector.pluralize('equipment') == 'equipment'
    assert inflector.pluralize('medium') == 'media'
    assert inflector.singularize('errata') == 'erratum'
